@@ -19,8 +19,8 @@
 """Progress reporting.
 
 This package provides progress reporting for the python-apt package. The module
-'base' provides classes with no output, the module 'gtk2' provides classes for
-GTK+ applications, and the module 'text' provides classes for terminals, etc.
+'base' provides classes with no output, and the module 'text' provides classes
+for terminals, etc.
 """
 
 from __future__ import print_function
