@@ -9,12 +9,13 @@ import apt_pkg
 import aptsources.sourceslist
 import aptsources.distro
 
+import testcommon
 
-class TestAptSources(unittest.TestCase):
+
+class TestAptSources(testcommon.TestCase):
 
     def setUp(self):
-        apt_pkg.init_config()
-        apt_pkg.init_system()
+        testcommon.TestCase.setUp(self)
         if apt_pkg.config["APT::Architecture"] not in ('i386', 'amd64'):
             apt_pkg.config.set("APT::Architecture", "i386")
         apt_pkg.config.set("Dir::Etc", os.getcwd())
@@ -26,6 +27,12 @@ class TestAptSources(unittest.TestCase):
         else:
             self.templates = "/usr/share/python-apt/templates/"
 
+    def tearDown(self):
+        aptsources.distro._OSRelease.OS_RELEASE_FILE = \
+            aptsources.distro._OSRelease.DEFAULT_OS_RELEASE_FILE
+        if "LSB_ETC_LSB_RELEASE" in os.environ:
+            del os.environ["LSB_ETC_LSB_RELEASE"]
+
     def testIsMirror(self):
         """aptsources: Test mirror detection."""
         yes = aptsources.sourceslist.is_mirror("http://archive.ubuntu.com",
@@ -247,6 +254,21 @@ class TestAptSources(unittest.TestCase):
         for key in found:
             self.assertEqual(found[key], 1)
 
+    @unittest.skip("lsb-release test broken when it was added")
+    def test_os_release_distribution(self):
+        """os-release file can be read and is_like is populated accordingly"""
+        os.environ["LSB_ETC_LSB_RELEASE"] = \
+            os.path.abspath("./data/aptsources/lsb-release")
+        aptsources.distro._OSRelease.OS_RELEASE_FILE = \
+            os.path.abspath("./data/aptsources/os-release")
+        distro = aptsources.distro.get_distro()
+        # Everything but is_like comes from lsb_release, see TODO in get_distro.
+        self.assertEqual('Ubuntu', distro.id)
+        self.assertEqual('xenial', distro.codename)
+        self.assertEqual('Ubuntu 16.04.1 LTS', distro.description)
+        self.assertEqual('16.04', distro.release)
+        self.assertEqual(['ubuntu', 'debian'], distro.is_like)
+
     def test_enable_disabled(self):
         """LP: #1042916: Test enabling disabled entry."""
         apt_pkg.config.set("Dir::Etc::sourcelist", "data/aptsources/"
